import copy
import logging
from typing import List, Set
import torch
from torch import nn

from torch_ema.torch_ema_utility import *


LOGGER = logging.getLogger(__name__)


class TorchEMA(nn.Module):
    """
    Implements exponential moving average.
    Utilizes an inverse decay schedule to manage longer term training runs.
    By adjusting the power, you can control how fast EMA will ramp up to your specified beta.
    
    Args:
        model (nn.Model):
            The model to wrap.
        beta (float):
            The exponential moving average. Default: 0.9999
        update_after_step (int):
            After this number of update calls it will start updating. Default: 100
        update_every (int):
            How often should be updated. Default: 10
        inv_gamma (float): 
            Inverse multiplicative factor of EMA warmup. Default: 1.
        power (float): 
            Exponential factor of EMA warmup. Default: 1.
        min_value (float): 
            The minimum EMA decay rate. Default: 0.
    """

    def __init__(
        self,
        model: nn.Module,
        beta: float = 0.9999,
        update_after_step: int = 100,
        update_every: int = 10,
        inv_gamma: float = 1.0,
        power: float = 2 / 3,
        min_value: float = 0.0,
        param_or_buffer_names_no_ema: Set = set()
    ) -> None:
        super(TorchEMA, self).__init__()
        self.beta: float = beta
        self.online_model: nn.Module = model

        try:
            self.ema_model: nn.Module = copy.deepcopy(model)
        except Exception as e:
            LOGGER.error(f"The model is not copyable: {e}")
            exit()

        self.ema_model.requires_grad_(False)
        self.update_every: int = update_every
        self.update_after_step: int = update_after_step

        self.inv_gamma: float = inv_gamma
        self.power: float = power
        self.min_value: float = min_value

        assert isinstance(param_or_buffer_names_no_ema, (set, list, Set, List))
        self.param_or_buffer_names_no_ema: Set = param_or_buffer_names_no_ema # parameter or buffer

        self.register_buffer('initted', torch.Tensor([False]))
        self.register_buffer('step', torch.tensor([0]))

    def restore_ema_model_device(self) -> None:
        self.ema_model.to(self.initted.device)

    def sync_params_between_models(self) -> None:
        for ema_params, current_params in zip(list(self.ema_model.parameters()), list(self.online_model.parameters())):
            if not is_float_dtype(current_params.dtype):
                continue

            ema_params.data.copy_(current_params.data)

        for ema_buffers, current_buffers in zip(list(self.ema_model.buffers()), list(self.online_model.buffers())):
            if not is_float_dtype(current_buffers.dtype):
                continue

            ema_buffers.data.copy_(current_buffers.data)

    def get_current_decay(self) -> float:
        epoch: int = clamp(self.step.item() - self.update_after_step - 1, min_value = 0)
        value: float = 1 - (1 + epoch / self.inv_gamma) ** - self.power

        if epoch <= 0:
            return 0.

        return clamp(value, min_value = self.min_value, max_value = self.beta)

    def update(self) -> None:
        step = self.step.item()
        self.step += 1

        if (step % self.update_every) != 0:
            return

        if step <= self.update_after_step:
            self.sync_params_between_models()
            return

        if not self.initted.item():
            self.sync_params_between_models()
            self.initted.data.copy_(torch.Tensor([True]))

        self.update_moving_average(self.ema_model, self.online_model)

    @torch.no_grad()
    def update_moving_average(self, ma_model, current_model) -> None:
        curr_decay = self.get_current_decay()

        for (name, curr_params), (_, ema_params) in zip(
            list(current_model.named_parameters()),
            list(ma_model.named_parameters())
        ):
            if not is_float_dtype(curr_params.dtype):
                continue

            if name in self.param_or_buffer_names_no_ema:
                ema_params.data.copy_(curr_params.data)
                continue

            diff = ema_params.data - curr_params.data
            diff.mul_(1.0 - curr_decay)
            ema_params.sub_(diff)

        for (name, curr_buffer), (_, ema_buffer) in zip(
            list(current_model.named_buffers()),
            list(ma_model.named_buffers())
        ):
            if not is_float_dtype(curr_buffer.dtype):
                continue

            if name in self.param_or_buffer_names_no_ema:
                ema_buffer.data.copy_(curr_buffer.data)
                continue

            diff = ema_buffer - curr_buffer
            diff.mul_(1.0 - curr_decay)
            ema_buffer.sub_(diff)

    def __call__(self, *args, **kwargs):
        return self.ema_model(*args, **kwargs)